
(in-package :pvs)

(defmacro make-xml-instance (class attrs &rest initargs)
  `(apply #'make-instance ,class
	  ,@initargs
	  ,attrs))

(defcl pvsid ()
  id place)

(defcl formals-elt ()
  formals)

(defcl assuming-part ()
  assumings)

(defun xml-parse (filename &optional output?)
  (let ((pvs-file (make-pathname :defaults filename :type "pvs"))
	(xml-file (make-pathname :defaults filename :type "xml")))
    (if (file-exists-p xml-file)
	(with-open-file (p xml-file)
	  (let ((lxml (net.xml.parser:parse-xml p :content-only p)))
	    (translate-xxml-to-xml lxml)))
	(pvs-error "XML error" (format nil "~a does not exist" xml-file)))))

(defun remove-empty-text-elements (lxml)
  (cond ((null lxml) nil)
	((consp lxml)
	 (mapcar #'remove-empty-text-elements
	   (remove-if #'empty-text-element? lxml)))
	(t lxml)))

(defun empty-text-element? (elt)
  (and (stringp elt)
       (every #'whitespace-char? elt)))

(defun whitespace-char? (c)
  (member c '(#\space #\tab #\newline #\page #\return #\linefeed)))

(defun translate-xxml-to-xml (lxml)
  (assert (null (cdr lxml)))
  (translate-xxml-to-xml* (remove-empty-text-elements (car lxml))))

(defun translate-xxml-to-xml* (lxml)
  (cond ((consp lxml)
	 (let ((fun (element-to-xml-instance (car lxml))))
	   (apply fun
	     (translate-xxml-attributes (car lxml))
	     (mapcar #'translate-xxml-to-xml* (cdr lxml)))))
	((stringp lxml)
	 (intern lxml))
	(t (break "translate-xxml-to-xml*: Not a list or string"))))

(defun element-to-xml-instance (lxml)
  (let ((eltname (if (listp lxml)
		     (if (listp (car lxml)) (caar lxml) (car lxml))
		     lxml)))
    (assert (symbolp eltname))
    (makesym "mk-xml-~(~a~)" eltname)))

(defun translate-xxml-attributes (elt)
  (when (listp elt)
    (translate-xxml-attributes* (cdr elt) nil)))

(defun translate-xxml-attributes* (list attrs)
  (if (null list)
      attrs
      (translate-xxml-attributes*
       (cddr list)
       (nconc attrs
	      (translate-xxml-attribute (car list) (cadr list))))))

(defun truep (value)
  (string-equal value "true"))

(defun translate-xxml-attribute (attr value)
  (case attr
    (place (list :place (translate-xxml-place value)))
    (ordnum (list :ordnum (parse-integer value :junk-allowed t)))
    (kind (list :kind value))
    (nonempty-type (list :nonempty-type (truep value)))
    (parens (list :parens (parse-integer value :junk-allowed t)))
    (infix (list :infix? (truep value)))
    (new-p (list :new-p (truep value)))
    (asserted-p (list :asserted-p (truep value)))
    (chain-p (list 'chain? (truep value)))
    (t (break "translate-xxml-attribute: ~a not handled" attr))))

(defun translate-xxml-place (string &optional (pos 0) place)
  (multiple-value-bind (num npos)
      (parse-integer string :start pos :junk-allowed t)
    (if (null num)
	(nreverse place)
	(translate-xxml-place string npos (cons num place)))))
       

(defmacro with-typecheck-context (xml-context &rest body)
  `(let* ((*current-theory* (get-xml-context ,xml-context))
	  (*current-context* (make-new-context *current-theory*))
	  (*generate-tccs* 'pvs::none))
     ,@body))


;;; Declarations

(defun mk-xml-theory (attrs id arg1 arg2 &rest args)
  (let* ((formals (when (formals-elt? arg1) (formals arg1)))
	 (assumings (if (assuming-part? arg1) arg1
			(when (assuming-part? arg2) arg2)))
	 (theory-decls (if (formals-elt? arg1)
			   (if (assuming-part? arg2)
			       (if (exporting? (car (last args)))
				   (butlast args)
				   args)
			       (if (exporting? (car (last args)))
				   (cons arg2 (butlast args))
				   (cons arg2 args)))
			   (if (assuming-part? arg2)
			       (if (exporting? (car (last args)))
				   (cons arg2 (butlast args))
				   (cons arg2 args))
			       (if (exporting? (car (last args)))
				   (cons arg1 (cons arg2 (butlast args)))
				   (cons arg1 (cons arg2 args))))))
	 (exporting (when (exporting? (car (last args)))
		      (car (last args))))
	 (theory (make-xml-instance 'module attrs
		   :id id
		   :formals formals
		   :assumings assumings
		   :theory theory-decls
		   :exporting exporting)))
    (mapcar #'(lambda (d) (when (declaration? d) (setf (module d) theory)))
      formals)
    (mapcar #'(lambda (d) (when (declaration? d) (setf (module d) theory)))
      assumings)
    (mapcar #'(lambda (d) (when (declaration? d) (setf (module d) theory)))
      theory-decls)
    theory))

(defun mk-xml-theory-id (attrs identifier)
  (mk-xml-id attrs identifier))

;;; exporting = attrs, exporting-names?, exporting-but-names?,
;;;             exporting-kind, exporting-theory-names?
(defun mk-xml-exporting (attrs &optional arg1 arg2 arg3 arg4)
  (let ((names (when (listp arg1) (break) arg1))
	(but-names (when (and (listp arg1) (listp arg2)) (break) arg2))
	(kind (if (listp arg1) (if (listp arg2) arg3) arg2) arg1)
	(theory-names (if (listp arg1) (if (listp arg2) arg4) arg4) arg2))
    (make-instance 'exporting
      :names names
      :but-names but-names
      :kind kind
      :modules theory-names)))

(defun mk-xml-exporting-kind (attrs kind)
  (mk-xml-id attrs kind))

;;;                    type-decl - -type-value, -nonempty 
;;;           nonempty-type-decl - -type-value, +nonempty
;;;                 type-eq-decl - +type-value, -type-from-decl, -nonempty
;;;        nonempty-type-eq-decl - +type-value, -type-from-decl, +nonempty

(defun mk-xml-type-decl (attrs pvsid
			       &optional formals type-value type-expr contains)
  (let ((nonempty? (truep (getf attrs 'nonempty)))
	(remattrs (remf attrs 'nonempty)))
    (if (null type-value)
	(make-xml-instance (if nonempty? 'nonempty-type-decl 'type-decl)
	    remattrs
	  :id (id pvsid)
	  :formals formals
	  :type-value type-value
	  :type-expr type-expr
	  :contains contains)
	(make-xml-instance (if nonempty? 'type-eq-decl 'nonempty-type-eq-decl)
	    remattrs
	  :id (id pvsid)
	  :formals formals
	  :type-value type-value
	  :type-expr type-expr
	  :contains contains))))

;;;               type-from-decl - +type-value, +type-from-decl, -nonempty
;;;      nonempty-type-from-decl - +type-value, +type-from-decl, +nonempty
;;;          struct-subtype-decl - +type-value, -type-struct-decl, -nonempty
;;; nonempty-struct-subtype-decl - +type-value, -type-struct-decl, +nonempty

(defun mk-xml-const-decl (attrs id arg1 &optional arg2 arg3)
  ;; Formals are optional, as is the definition
  (let ((formals (when (listp arg1) arg1))
	(type (if (listp arg1) arg2 arg1))
	(definition (if (listp arg1) arg3 arg2)))
    (make-xml-instance 'const-decl attrs
      :id (id id)
      :declared-type type
      :definition definition)))

(defun mk-xml-typedeclaration (attrs identifier &optional typedef)
  (when (typep typedef '(or datatype scalar-type))
    (setf (id typedef) (id identifier)))
  (make-xml-instance 'type-declaration attrs
		     'id (id identifier)
		     'type-expr typedef))

(defmethod id ((x symbol))
  x)

(defmethod mk-xml-actual ((ex expr))
  (make-instance 'pvs:actual
    :expr ex))

(defmethod mk-xml-actual ((ex type-expr))
  (make-instance 'pvs:actual
    :expr ex
    ;;:type-value ex
    ))
    

;;; Types

(defun mk-xml-scalartype (attrs &rest scalarelements)
  (make-xml-instance 'scalar-type attrs
    'identifiers (mapcar #'id scalarelements)))

(defun mk-xml-scalarelement (attrs id)
  (make-xml-instance 'scalar-element attrs 'id id))

(defun mk-xml-datatype (attrs &rest constructors)
  (make-xml-instance 'datatype attrs
		     'constructors constructors))

(defun mk-xml-constructor (attrs identifier &rest accessors)
  (make-xml-instance 'constructor attrs
		     'id (id identifier)
		     'arguments accessors
		     'recognizer (intern (format nil "~a?" (id identifier)))))

(defun mk-xml-accessor (attrs identifier type)
  (make-xml-instance 'accessor attrs
		     'id (id identifier)
		     'declared-type type))

(defun mk-xml-type-name (attrs id
			       &optional theory-id library actuals
			       mappings target)
  (assert (and id (symbolp id)))
  (make-xml-instance 'type-name attrs
		     :id id
		     :mod-id theory-id
		     :library library
		     :actuals actuals
		     :mappings mappings
		     :target target))

(defun mk-xml-subrange (attrs low high)
  (make-xml-instance 'subrange attrs
		     'lower-bound low
		     'upper-bound high))

(defun mk-xml-unbounded ()
  (break))

(defun mk-xml-arraytype (attrs indextype type)
  (make-xml-instance 'array-type attrs
		     'domain indextype
		     'range type))

(defun mk-xml-tuple-type (attrs &rest types)
  (assert (every #'type-expr? types))
  (make-xml-instance 'tupletype attrs
    :types types))

(defun mk-xml-recordtype (attrs &rest fields)
  (make-xml-instance 'record-type attrs
		     'fields fields))

(defun mk-xml-fielddeclaration (attrs id type)
  (make-xml-instance 'field-declaration attrs
		     'id id
		     'declared-type type))

(defun mk-xml-function-type (attrs domain range)
  (assert (and (type-expr? domain) (type-expr? range)))
  (make-xml-instance 'funtype attrs
    :domain domain
    :range range))

(defun mk-xml-nameexpr (attrs id)
  (assert (symbolp id))
  (make-xml-instance 'name-expr attrs
		     'id id))

(defun mk-xml-qualifiednameexpr (attrs identifier context-name)
  (make-xml-instance 'qualified-name-expr attrs
		     'id (id identifier)
		     'context-name context-name))

(defun mk-xml-application (attrs operator argument)
  (let ((infix? (getf attrs 'infix?))
	(unary? (getf attrs 'unary?)))
    (remf attrs 'infix?)
    (remf attrs 'unary?)
    (make-xml-instance (xml-application-class operator infix? unary?)
	attrs
      'operator operator
      'argument (if (or (not (tuple-literal? argument))
			(cdr (exprs argument)))
		    argument
		    (car (exprs argument))))))

(defun xml-application-class (operator infix? unary?)
  (case operator
    (NOT (if unary? 'unary-negation 'negation))
    (AND (if infix? 'infix-conjunction 'conjunction))
    (OR  (if infix? 'infix-disjunction 'disjunction))
    ;;(XOR (if infix? 'infix-xor-expr 'xor-expr))
    (=>  (if infix? 'infix-implication 'implication))
    (<=> (if infix? 'infix-iff 'iff))
    (=   (if infix? 'infix-equation 'equation))
    (/=  (if infix? 'infix-disequation 'disequation))
    (t   (cond (infix? 'infix-application)
	       (unary? 'unary-application)
	       (t 'application)))))

(defun mk-xml-recordselection (attrs record field)
  (make-xml-instance 'record-selection attrs
    'argument record
    'id (id field)))

(defun mk-xml-tupleselection (attrs tuple index)
  (make-xml-instance 'tuple-selection attrs
    'argument tuple
    'index (number index)))

(defun mk-xml-recordliteral (attrs &rest entries)
  (make-xml-instance 'record-literal attrs
		     'assignments entries))

(defun mk-xml-recordentry (attrs id expr)
  (make-xml-instance 'record-entry attrs
		     'id (id id)
		     'expression expr))

(defun mk-xml-tupleliteral (attrs &rest exprs)
  (make-xml-instance 'tuple-literal attrs
		     'exprs exprs))

(defun mk-xml-updateexpression (attrs depth updateposition value)
  (make-xml-instance 'update-expression attrs
		     'expression (update-operator updateposition depth)
		     'updateposition updateposition
		     'value value))

(defun mk-xml-depth (attrs depth)
  (parse-integer (string depth)))

(defun mk-xml-arrayposition ()
  (break))

(defun mk-xml-arrayliteral (attrs binding expr)
  (make-xml-instance 'array-literal attrs
		     'bindings (list binding)
		     'expression expr))

(defun mk-xml-indexvardecl (attrs id type)
  (make-xml-instance 'index-var-decl attrs
		     'id (id id)
		     'declared-type type))

(defun mk-xml-lambdaabstraction (attrs bindings expr)
  (make-xml-instance 'lambda-abstraction attrs
    'bindings (mapcar #'(lambda (b)
			  (change-class b 'bind-decl))
		(if (var-decls? bindings)
		    (declarations bindings)
		    bindings))
    'expression expr))

(defun mk-xml-quantifiedexpression (attrs quant bindings expr)
  (make-xml-instance 'quantified-expression attrs
		     'quantifier quant
		     'bindings (mapcar #'(lambda (b)
					   (change-class b 'bind-decl))
				 (if (var-decls? bindings)
				     (declarations bindings)
				     bindings))
		     'expression expr))

(defun mk-xml-quantifier (attrs quant)
  (declare (ignore attrs))
  quant)

(defun mk-xml-letexpression (attrs declarations expression)
  (make-xml-instance 'let-expression attrs
		     'declarations declarations
		     'expr expression))

(defun mk-xml-letdeclarations (attrs &rest let-decls)
  (declare (ignore attrs))
  let-decls)

(defun mk-xml-letdeclaration (attrs id type value)
  (make-xml-instance 'let-declaration attrs
		     'id (id id)
		     'declared-type type
		     'value value))

(defun mk-xml-setpredexpression (attrs id type pred)
  (make-xml-instance 'set-pred-expr attrs
		     'bindings (list (make-xml-instance 'var-decl nil
							'id (id id)
							'declared-type type))
		     'expression pred))

(defun mk-xml-setlistexpression (attrs &rest exprs)
  (make-xml-instance 'set-list-expr attrs
		     'expressions exprs))

(defun mk-xml-conditional (attrs cond then else)
  (let ((elsif? (getf attrs 'elsif?)))
    (remf attrs 'elsif?)
    (make-xml-instance (if elsif?
			   'chained-conditional
			   'conditional)
	attrs
      'operator (make-xml-instance 'name-expr nil 'id 'IF)
      'argument (make-xml-instance 'tuple-literal nil
		  'exprs (list cond then else)))))

(defun mk-xml-numeral (attrs num)
  (make-xml-instance 'numeral attrs
		     'number (parse-integer (string num))))

(defun mk-xml-vardecl (attrs id type)
  (make-xml-instance 'var-decl attrs
		     'id (id id)
		     'declared-type type))

(defun mk-xml-typedecl (attrs identifier)
  (make-xml-instance 'formal-type-decl attrs
		     'id (id identifier)))

(defun mk-xml-identifiers (attrs &rest ids)
  (declare (ignore attrs))
  ids)

(defun mk-xml-id (attrs id)
  (if (stringp id) (intern id :pvs) id))
